/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/

/* This file, resp.c, is adapted from original code by Steve Kirkendall.
Permission given. */

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include "client.h"

/* handle a message from the host */
int doresp ( SERVERM_UL *resp ) {
	SIMPLECLIENT_UL	*cli, *temp;
	CLIENTM_UL	req;
	FILE		*fp;
	int		i;
	struct stat	filestat;
	char		buffer[100];

	switch (resp->response) {
	  case UL_WELCOME:
		fprintf( stderr, "Received UL_WELCOME in doresp, this shouldn't happen. Ignoring.\n" );
		return OK;

	  case UL_JOINED:
		for ( i = 1, temp = firstcli; temp->next != NULL; temp = temp->next, i++ ) ;

		if ( (cli = (SIMPLECLIENT_UL *)malloc(sizeof(SIMPLECLIENT_UL)))
				== NULL ) {
			fprintf( stderr, "Couldn't allocate memory in doresp: UL_JOINED.\n" );
			return FATALERROR;
		}

		temp->next = cli;
		cli->status	= READY;
		cli->mass	= resp->joinrace.mass;
		cli->racepos	= i;	/*** temporary? */
		cli->id		= (resp->joinrace.number<<16);
#ifdef CLIENT
		memcpy( &cli->color, &resp->joinrace.color, sizeof(ggi_color));
#endif
		cli->model	= strdup( resp->joinrace.model );
		cli->driver	= strdup( resp->joinrace.driver );
		readGeometry( cli->model, cli->geom );
		cli->ai		= resp->joinrace.ai;
		cli->next	= NULL;
		/**** should calculate position */
#ifdef CLIENT
		sprintf( buffer, "%s (%d) joined race.", resp->joinrace.driver,
				resp->joinrace.number );
		addmessage( buffer, -1 );
#endif
		return OK;

	  case UL_LEFT:
		for ( temp = firstcli; (temp->next->id>>16) != resp->leftrace.number;
				temp = temp->next) ;
		cli = temp->next;
#ifdef CLIENT
		sprintf( buffer, "%s (%d) left race.", cli->driver,
				cli->id>>16 );
		addmessage( buffer, -1 );
#endif
		free( cli->model );
		free( cli->driver );
		freeDoodad( cli->geom );
		temp->next = cli->next;
		free( cli );
		return OK;

	  case UL_COUNTDOWN:
		/****** do something */
		return OK;

	  case UL_ENDRACE:
#ifdef CLIENT
		addmessage( "Race is over!", -1 );
#endif
		/*** show final grid */
		return FATALERROR;

	  case UL_SMESSAGE:
#ifdef CLIENT
		addmessage( resp->smessage.message, 0 );
#endif
		return OK;

	  case UL_CMESSAGE:
#ifdef CLIENT
		addmessage( resp->cmessage.message, resp->cmessage.from );
#endif
		return OK;

	  case UL_PAUSE:
		paused = True;
		return OK;

	  case UL_UNPAUSE:
		paused = False;
		return OK;

	/* note to 'smart hackers': even if you ignore UL_QUIT messages and
	   don't reply to the server, the server will ignore any messages
	   from you. So, be polite and don't erase this. Go hack some other
	   part of the code! */
	  case UL_QUIT: /* server kicked us out! */
		req.request = UL_QUIT;
		netclientsend( &req );
		netflush();
		return FATALERROR;

	  case UL_CONTROL:
		for ( temp = firstcli; (temp->id>>16) != resp->control.number;
				temp = temp->next) ;
		temp->position.x += ((double)resp->control.dx)*0.01;
		temp->position.y += ((double)resp->control.dy)*0.01;
		temp->position.z += ((double)resp->control.dz)*0.01;
		temp->direction.x += ((double)resp->control.dwx)*0.01;
		temp->direction.y += ((double)resp->control.dwy)*0.01;
		temp->direction.z += ((double)resp->control.dwz)*0.01;
		temp->angle += ((double)resp->control.dangle)*0.01;
		/***** what should we do with timestamp??? */
		return OK;

	  case UL_PRECISECONTROL:
		for ( temp = firstcli; (temp->id>>16) != resp->precisecontrol.number;
				temp = temp->next) ;
		copyV( &temp->position, &resp->precisecontrol.position );
		copyV( &temp->direction, &resp->precisecontrol.direction );
		temp->angle = resp->precisecontrol.angle;
		/***** what should we do with timestamp??? */
		return OK;

	  case UL_GETFILE:
		req.request = UL_FILE;
		req.file.name = resp->getfile.name;
		/* we're supposing that name has the subdirectory. */
		if ( strstr( resp->getfile.name, ".." ) ) {
			/* trying to get some file outside of ULDIR. This is
			a severe security bug if allowed */
			fprintf( stderr, "%s%s%s", ULDIR, resp->getfile.name,
				" asked; access denied!\n" );
			return OK;
		}
		sprintf( buffer, "%s%s", ULDIR, resp->getfile.name );
		if ( stat( buffer, &filestat ) != 0 || 
			fopen( buffer, "r" ) == NULL ) {
			fprintf( stderr, "Could not get stat of %s\n",
					 resp->getfile.name );
			req.file.size = 0;
			req.file.data = NULL;
			netclientsend( &req );
			netflush();
			return OK;	/* ? */
		}
		req.file.size = (int)filestat.st_size;
		/* there's no space for '\0' because files are treated specially
		   - only req.file.size bytes are read */
		/*** if we could send the file without having to put it in
		 memory it'd be much better */
		req.file.data = (unsigned char *)malloc(req.file.size);
		fread( req.file.data, req.file.size, 1, fp );
		netclientsend( &req );
		netflush();
		free( req.file.data );
		return OK;

	  case UL_FILE:
		sprintf( buffer, "%s%s", ULDIR, resp->getfile.name );
		if ( resp->file.data == 0 ) {
			fprintf( stderr, "No data sent in %s. This is a server error.\n", 
					buffer );
			free( resp->file.data );
			return OK; /* or error? */
		}
		if ( !(fp = fopen( buffer, "w" )) ) {
			fprintf( stderr, "Couldn't save file %s!\n", 
					buffer );
			free( resp->file.data );
			return OK; /* or error? */
		}
		if ( fwrite( resp->file.data, resp->file.size, 1, fp ) != 
				resp->file.size ) {
			fprintf( stderr, "Couldn't write to file %s!\n", 
					buffer );
			fclose( fp );
			free( resp->file.data );
			return OK; /* or error? */
		}
		fclose( fp );
		free( resp->file.data );
		return OK;

	  case UL_COLLISION:
		/* emit sound */
		/* make checks */
		break;

	  case UL_SPECIAL:
		if ( own->message )
			own->message( &req, resp );
		break;

#ifdef OLD_CODE_FROM_XRACE_IGNORE_IT
	  case UL_COUNTDOWN:
		if (resp->countdown.time > 0)
		{
			servermode = ULMODE_COUNTDOWN;
			sprintf(center, "{%ld...}", resp->countdown.time / 1000L);
			for (v = view; v; v = v->next)
			{
				v->center = center;
				v->hidecenter = now + 1000;
			}
			soundevent("countdown");
		}
		else
		{
			servermode = ULMODE_RACE;
			for (v = view; v; v = v->next)
			{
				v->center = "{go!}";
				v->hidecenter = now + 1000;
			}
			soundevent("start");
		}
		break;
#endif
	}
}

/* this message is provided for security and anti-cheating */
void sendSpecialMessageR ( CLIENTM_UL *request ) {
	SIMPLECLIENT_UL	*tmp;
	int		i = 0;

	/* make sure it's a special message */
	request->request = UL_SPECIAL;

	/* make sure the message isn't spoofed */
	request->special.from = own->id>>16;
	
	/* check if the target exists */
	if ( request->special.to != 0 )
		for ( tmp = firstcli; tmp != NULL; tmp = tmp->next )
			i += ( (tmp->id>>16) == abs(request->special.to) ? 1 : 0);

	if ( i == 0 ) {
		fprintf( stderr, "Target of special message doesn't exist\n" );
		return;
	}
	netclientsend( request );
}
