/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
#include "mathlib.h"

void initQuatD ( QUATERNION_UL *q, double s, double x, double y, double z ) {
	q->s = s;
	q->v.x = x;
	q->v.y = y;
	q->v.z = z;
}

void initQuatVS ( QUATERNION_UL *q, double s, VECTOR_UL *v ) {
	q->s = s;
	q->v.x = v->x;
	q->v.y = v->y;
	q->v.z = v->z;
}

/* returns normalized */
void initQuatVA ( QUATERNION_UL *q, double angle, VECTOR_UL *v ) {
	double cosangle, sinangle;
	copyV( &q->v, v );
	normalizeV( &q->v );
	cosangle = cos( angle/2.0 );
	sinangle = sin( angle/2.0 );
	q->s = cosangle;
	scalarMV( &q->v, sinangle, &q->v );
}

void zeroQ ( QUATERNION_UL *q ) {
	q->s = q->v.x = q->v.y = q->v.z = 0.0;
}	

void copyQ ( QUATERNION_UL *dest, QUATERNION_UL *src ) {
	memcpy( dest, src, sizeof(QUATERNION_UL) );
}

double normQ ( QUATERNION_UL *q ) {
	return sqrt( q->s*q->s + q->v.x*q->v.x + q->v.y*q->v.y + q->v.z*q->v.z );
}

void normalizeQ ( QUATERNION_UL *q, QUATERNION_UL *r ) {
	double n = normQ( q );
	r->s = q->s/n;
	r->v.x = r->v.x/n;
	r->v.y = r->v.y/n;
	r->v.z = r->v.z/n;
}

void conjugateQ ( QUATERNION_UL *q, QUATERNION_UL *r ) {
	r->s = q->s;
	r->v.x = -q->v.x;
	r->v.y = -q->v.y;
	r->v.z = -q->v.z;
}

void addQ ( QUATERNION_UL *q1, QUATERNION_UL *q2, QUATERNION_UL *r ) {
	r->s = q1->s + q2->s;
	r->v.x = q1->v.x + q2->v.x;
	r->v.y = q1->v.y + q2->v.y;
	r->v.z = q1->v.z + q2->v.z;
}

void subQ ( QUATERNION_UL *q1, QUATERNION_UL *q2, QUATERNION_UL *r ) {
	r->s = q1->s - q2->s;
	r->v.x = q1->v.x - q2->v.x;
	r->v.y = q1->v.y - q2->v.y;
	r->v.z = q1->v.z - q2->v.z;
}

void scalarMQ( QUATERNION_UL *q, const double m, QUATERNION_UL *r) {
	r->s = m*q->s;
	r->v.x = m*q->v.x;
	r->v.y = m*q->v.y;
	r->v.z = m*q->v.z;
}

void multQ ( QUATERNION_UL *q1, QUATERNION_UL *q2, QUATERNION_UL *r ) {
	QUATERNION_UL t;	
	t.s = q1->s*q2->s - q1->v.x*q2->v.x - q1->v.y*q2->v.y - q1->v.z*q2->v.z;
	t.v.x = q1->s*q2->v.x + q1->v.x*q2->s + q1->v.y*q2->v.z - q1->v.z*q2->v.y;
	t.v.y = q1->s*q2->v.y - q1->v.x*q2->v.z + q1->v.y*q2->s + q1->v.z*q2->v.x;
	t.v.z = q1->s*q2->v.z + q1->v.x*q2->v.y - q1->v.y*q2->v.x + q1->v.z*q2->s;
	copyQ( r, &t );
}

void inverseQ ( QUATERNION_UL *q1, QUATERNION_UL *r ) {
	conjugateQ( q1, r );
	scalarMQ( r, 1/normQ( q1 ), r );
}

/* BUG? is it correct? */
void qvq ( QUATERNION_UL *q1, VECTOR_UL *v, QUATERNION_UL *q2, VECTOR_UL *r ) {
	QUATERNION_UL p;

	copyV( &p.v, v );
	p.s = 0;
	multQ( q1, &p, &p );
	multQ( &p, q2, &p );
	copyV( r, &(p.v) );
}
