/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/
/* This file, netio.c, is adapted from original code by Steve Kirkendall.
Permission given. */

#include <stdio.h>
#include "unlimited.h"
#include "server.h"

#define SENDQ	16384
#define RECVQ	16384

static int sendfd;		/* fd of server to send to, or -1 for all */
static char sendq[16384];	/* output queue */
static int sendcnt;		/* number of bytes in output queue */


/* Add a 16-bit int value to sendq[] */
static void sendshort ( short l ) {
	sendq[sendcnt++] = l >> 8;
	sendq[sendcnt++] = l;
}

/* Add a 32-bit int value to sendq[] */
static void sendlong ( long l ) {
	sendq[sendcnt++] = l >> 24;
	sendq[sendcnt++] = l >> 16;
	sendq[sendcnt++] = l >> 8;
	sendq[sendcnt++] = l;
}

/* Add a string to sendq[] */
static void sendstring ( char *s ) {
	int	i;

	/* Default string is always "" */
	if (!s)
		s = "";

	/* Add the string to the sendq[].  May need to flush the buffer */
	i = strlen(s) + 1;
	if ( sendcnt + i > SENDQ ) {
		netflush();
		while (i > SENDQ - 1024) {
			sendcnt = (i < SENDQ ? i : SENDQ);
			memcpy(sendq, s, sendcnt);
			s += sendcnt;
			i -= sendcnt;
			netflush();
		}
		if (i == 0)
			return;
	}
	memcpy( &sendq[sendcnt], s, i );
	sendcnt += i;
}


/* Send a response to a given client or clients.  If fd is positive, the
 * message will be sent to the client with that fd.  If it is 0, it will be
 * sent to all clients.  If negative, it will be sent to all except the
 * one whose fd is the opposite of the given fd.
 */
void netsend ( int fd, SERVERM_UL *msg ) {
	int	i;

	/* if sendq is nearly full, or this msg is to a different client,
	 * then send it now
	 */
	if (fd != sendfd || sendcnt > SENDQ - 1024)
		netflush();
	sendfd = fd;

	/* stuff the request type into the send queue */
	sendq[sendcnt++] = msg->response;

	/* convert the request to binary */
	switch (msg->response) {
	  case UL_WELCOME:
		sendshort( msg->welcome.number );
		sendshort( msg->welcome.racersn );
		sendstring( msg->welcome.track );
		sendstring( msg->welcome.models );
		sendstring( msg->welcome.drivers );
		sendstring( msg->welcome.mass );
		sendstring( msg->welcome.numbers );
		sendstring( msg->welcome.color );
		break;

	  case UL_JOINED:
		sendstring( msg->joinrace.model );
		sendstring( msg->joinrace.driver );
		sendlong( (long)msg->joinrace.mass*10 );
		sendshort( msg->joinrace.number );
		sendstring( msg->joinrace.color );
		sendq[sendcnt++] = msg->joinrace.ai;		
		break;

	  case UL_LEFT:
		sendshort( msg->joinrace.number );
		break;

	  case UL_COUNTDOWN:
		sendshort( msg->countdown.time );
		break;

	  case UL_SMESSAGE:
		sendshort( msg->cmessage.to );
		sendstring( msg->smessage.message );
		break;

	  case UL_CMESSAGE:
		sendshort( msg->cmessage.from );
		sendshort( msg->cmessage.to );
		sendstring( msg->cmessage.message );
		break;

	  case UL_PAUSE:
	  case UL_UNPAUSE:
	  case UL_QUIT:
	  case UL_ENDRACE:
		break;

	  case UL_CONTROL:
		sendshort( msg->control.number );
		sendshort( msg->control.dx );
		sendshort( msg->control.dy );
		sendshort( msg->control.dz );
		sendshort( msg->control.dwx );
		sendshort( msg->control.dwy );
		sendshort( msg->control.dwz );
		sendshort( msg->control.dangle );
		sendshort( msg->control.timestamp );
		break;

	  case UL_PRECISECONTROL:
		sendshort( msg->precisecontrol.number );
		sendlong( (long)(msg->precisecontrol.position.x*1000) );
		sendlong( (long)(msg->precisecontrol.position.y*1000) );
		sendlong( (long)(msg->precisecontrol.position.z*1000) );
		sendlong( (long)(msg->precisecontrol.direction.x*1000) );
		sendlong( (long)(msg->precisecontrol.direction.y*1000) );
		sendlong( (long)(msg->precisecontrol.direction.z*1000) );
		sendlong( (long)(msg->precisecontrol.angle*1000) );
		sendshort( msg->precisecontrol.timestamp );
		break;
		
	  case UL_GETFILE:
		sendstring( msg->getfile.name );
		break;

	  case UL_FILE:
		sendstring( msg->file.name );
		sendlong( msg->file.size );
		sendstring( msg->file.data );
		break;
		
	  case UL_COLLISION:
		sendlong( msg->collision.id );
		sendlong( (long)(msg->collision.p.x*1000) );
		sendlong( (long)(msg->collision.p.y*1000) );
		sendlong( (long)(msg->collision.p.z*1000) );
		sendshort( msg->collision.timestamp );
		break;
		
	  case UL_OBJECTCREATE:
		sendstring( msg->objectcreate.name );
		sendlong( msg->objectcreate.id );
		sendlong( (long)(msg->objectcreate.position.x*1000) );
		sendlong( (long)(msg->objectcreate.position.y*1000) );
		sendlong( (long)(msg->objectcreate.position.z*1000) );
		sendlong( (long)(msg->objectcreate.direction.x*1000) );
		sendlong( (long)(msg->objectcreate.direction.y*1000) );
		sendlong( (long)(msg->objectcreate.direction.z*1000) );
		sendlong( (long)(msg->objectcreate.angle*1000) );
		sendlong( (long)(msg->objectcreate.mass*1000) );
		sendshort( msg->objectcreate.timestamp );
		break;
		
	  case UL_OBJECTDESTROY:
		sendlong( msg->objectdestroy.id );
		sendlong( (long)(msg->objectdestroy.position.x*1000) );
		sendlong( (long)(msg->objectdestroy.position.y*1000) );
		sendlong( (long)(msg->objectdestroy.position.z*1000) );
		sendq[sendcnt++] = msg->objectdestroy.causamortis;
		sendshort( msg->objectdestroy.timestamp );
		break;

	  case UL_OBJECTMODIFY:
		sendlong( msg->objectmodify.id );
		sendlong( (long)(msg->objectmodify.position.x*1000) );
		sendlong( (long)(msg->objectmodify.position.y*1000) );
		sendlong( (long)(msg->objectmodify.position.z*1000) );
		sendlong( (long)(msg->objectmodify.direction.x*1000) );
		sendlong( (long)(msg->objectmodify.direction.y*1000) );
		sendlong( (long)(msg->objectmodify.direction.z*1000) );
		sendlong( (long)(msg->objectmodify.angle*1000) );
		sendshort( msg->objectmodify.timestamp );
		break;

	  case UL_OBJECTMOVE:
		sendlong( msg->objectmove.id );
		sendshort( msg->objectmove.dx );
		sendshort( msg->objectmove.dy );
		sendshort( msg->objectmove.dz );
		sendshort( msg->objectmove.dwx );
		sendshort( msg->objectmove.dwy );
		sendshort( msg->objectmove.dwz );
		sendshort( msg->objectmove.dangle );
		sendshort( msg->objectmove.timestamp );
		break;
		
	  case UL_SPECIAL:
		sendshort( msg->special.from );
		sendshort( msg->special.to );
		sendq[sendcnt++] = msg->special.type;
		sendstring( msg->special.data );
		break;

	  default:
		fprintf(stderr, "invalid response %d\n", msg->response);
	}
}

/* flush outgoing messages, which were queued by netsend(). */
void netflush ( void ) {
	SIMPLECLIENT_UL	*client;

	if ( sendcnt > 0 ) {
		if ( sendfd > 0 )
			(void)netwrite( sendfd, sendq, sendcnt );
		else {
			for ( client = firstclient; client; client = client->next ) {
				if ( client->sock != -sendfd
				 && netwrite( client->sock, sendq, sendcnt ) != sendcnt )
				 	perror("server: netwrite");
			}
		}
		sendcnt = 0;
	}
}


/*---------------------------------------------------------------------------*/

static char 	recvq[RECVQ];
static int	recvcnt;
static int	recvused;

/* Test whether all of the required fields are present.  Return 1 if missing,
 * or 0 if complete.
 */
static int incomplete ( int bytes, int strings ) {
	/* is it obviously too short? */
	if (recvused + bytes + strings > recvcnt)
		return 1;

	/* if no strings needed, then we got it */
	if (strings == 0)
		return 0;

	/* count the strings */
	for (; recvused + bytes < recvcnt; bytes++)
		if (recvq[recvused + bytes] == '\0' && --strings == 0)
			return 0;

	/* ran out of bytes before the last string */
	return 1;
}


static short recvshort ( void ) {
	short	s;

	s = (((short)recvq[recvused] << 8) & 0xff00)
	   | (((short)recvq[recvused + 1]) & 0xff);
	recvused += 2;
	return s;
}

static long recvlong ( void ) {
	long	l;

	l = ((((long)recvq[recvused] << 24) & 0xff000000)
	   | (((long)recvq[recvused + 1] << 16) & 0xff0000)
	   | (((long)recvq[recvused + 2] << 8) & 0xff00)
	   | (((long)recvq[recvused + 3] & 0xff)));
	recvused += 4;
	return l;
}


/* Receive a message from a particular client, identified by its fd.  It is
 * assumed that a message is known to be waiting; if not, this function will
 * block until one becomes available.
 *
 * Returns a CLIENTM_UL struct, or NULL if there are no messages (connection
 * to client was lost).  Any (char *) fields in the response will point to
 * statically allocated space --  you don't need to free it, but you should
 * make your own copy of it before the next call to netrcv().
 *
 * After successfully reading from a client, you *must* continue to read from
 * that client as long as netmore() returns TRUE.
 */
CLIENTM_UL *netrcv ( int fd ) {
	static CLIENTM_UL request;
	SERVERM_UL	resp;
	int		beginning;
	int		i;

	/* if no bytes in recvq, then read some */
	if (recvused == recvcnt) {
		recvcnt = netread(fd, recvq, RECVQ);
		if (recvcnt <= 0)
			return NULL;
		recvused = 0;
	}

	/* We may have an incomplete message.  Do some preparation so we can
	 * read some more and start over if necessary.
	 */
	beginning = recvused;

	for (;;) {
		/* parse the request */
		request.request = recvq[recvused++];
		switch ( request.request ) {
		  case UL_HELLO:
			if ( incomplete( 7, 2 ) )
				break;
			request.hello.ai = recvq[recvused++];
			request.hello.driver = &recvq[recvused];
			while ( recvq[recvused++] ) { }
			request.hello.model = &recvq[recvused];
			while (recvq[recvused++]) { }
			request.hello.color.r = recvshort();
			request.hello.color.g = recvshort();
			request.hello.color.b = recvshort();
			return &request;

		  case UL_QUIT:
		  case UL_PAUSE:
		  case UL_UNPAUSE:
		  	/* no parameters */
			return &request;

		  case UL_CMESSAGE:
			if ( incomplete( 4, 1 ) )
				break;
			request.cmessage.from = (unsigned short)recvshort();
			request.cmessage.to = (unsigned short)recvshort();
			request.cmessage.message = &recvq[recvused];
			while (recvq[recvused++]) { }
			return &request;

		  case UL_FILE:
			if ( incomplete( 4, 1 ) ) {
#ifdef UNTOUCHED
				/* It might not fit in recvq[].  We'd better
				 * receive the string piecemeal, and assemble
				 * it using nrbuild().
				 */
				(void)nrbuild(NULL);
				do {
					request.gettrack.data =
						nrbuildn(&recvq[recvused],
							recvcnt-recvused);
					recvcnt = nrread(fd, recvq, RECVQ);
					if (recvcnt <= 0)
						return NULL;
					recvused = 0;
				} while (incomplete(0, 1));
				i = strlen(&recvq[recvused]) + 1;
				request.track.data = nrbuildn(&recvq[recvused], i);
				recvused += i;
#endif
			}
			else {
				/* We got lucky.  It fits */
				request.file.name = &recvq[recvused];
				while ( recvq[recvused++] ) { }
				request.file.size = (unsigned long)recvlong();
				request.file.data = &recvq[recvused];
				while ( recvq[recvused++] ) { }
			}
			return &request;

		  case UL_GETFILE:
			if (incomplete(0, 1))
				break;
			request.getfile.name = &recvq[recvused];
			while (recvq[recvused++]) { }
			return &request;
			
		  case UL_CONTROL:
			if ( incomplete( 16, 0 ) ) 
				break;
			request.control.dx = recvshort();
			request.control.dy = recvshort();
			request.control.dz = recvshort();
			request.control.dwx = recvshort();
			request.control.dwy = recvshort();
			request.control.dwz = recvshort();
			request.control.dangle = recvshort();
			request.control.timestamp = recvshort();
			return &request;

		  case UL_PRECISECONTROL:
			if ( incomplete( 30, 0 ) ) 
				break;
			request.precisecontrol.position.x = ((double)recvlong())*0.001;
			request.precisecontrol.position.y = ((double)recvlong())*0.001;
			request.precisecontrol.position.z = ((double)recvlong())*0.001;
			request.precisecontrol.direction.x = ((double)recvlong())*0.001;
			request.precisecontrol.direction.y = ((double)recvlong())*0.001;
			request.precisecontrol.direction.z = ((double)recvlong())*0.001;
			request.precisecontrol.angle = ((double)recvlong())*0.001;
			request.precisecontrol.timestamp = recvshort();
			return &request;

		  case UL_COLLISION:
			request.collision.number = recvlong();
			request.collision.p.x = ((double)recvlong())*0.001;
			request.collision.p.y = ((double)recvlong())*0.001;
			request.collision.p.z = ((double)recvlong())*0.001;
			request.collision.timestamp = recvshort();
			return &request;

		  case UL_OBJECTCREATE:
			if ( incomplete( 38, 1 ) ) 
				break;
			if ( general.createobject == False ) {
				resp.response = UL_SMESSAGE;
				resp.smessage.to = fd;
				(const)resp.smessage.message = noobjcreate;
				netsend( fd, &resp );
				request.request = UL_FOO;
			}
			request.objectcreate.name = &recvq[recvused];
			while (recvq[recvused++]) { }
			request.objectcreate.id = recvlong();
			request.objectcreate.position.x = ((double)recvlong())*0.001;
			request.objectcreate.position.y = ((double)recvlong())*0.001;
			request.objectcreate.position.z = ((double)recvlong())*0.001;
			request.objectcreate.direction.x = ((double)recvlong())*0.001;
			request.objectcreate.direction.y = ((double)recvlong())*0.001;
			request.objectcreate.direction.z = ((double)recvlong())*0.001;
			request.objectcreate.angle = ((double)recvlong())*0.001;
			request.objectcreate.mass = ((float)recvlong())*0.001;
			request.objectcreate.timestamp = recvshort();
			return &request;

		  case UL_OBJECTDESTROY:
			if ( incomplete( 19, 0 ) ) 
				break;
			if ( general.destroyobject == False ) {
				resp.response = UL_SMESSAGE;
				resp.smessage.to = fd;
				(const)resp.smessage.message = noobjdestroy;
				netsend( fd, &resp );
				request.request = UL_FOO;
			}
			request.objectdestroy.id = recvlong();
			request.objectdestroy.position.x = ((double)recvlong())*0.001;
			request.objectdestroy.position.y = ((double)recvlong())*0.001;
			request.objectdestroy.position.z = ((double)recvlong())*0.001;
			request.objectdestroy.causamortis = recvq[recvused++];
			request.objectdestroy.timestamp = recvshort();
			return &request;

		  case UL_OBJECTMODIFY:
			if ( incomplete( 30, 0 ) ) 
				break;
			if ( general.modifyobject == False ) {
				resp.response = UL_SMESSAGE;
				resp.smessage.to = fd;
				(const)resp.smessage.message = noobjmodify;
				netsend( fd, &resp );
				request.request = UL_FOO;
			}
			request.objectmodify.position.x = ((double)recvlong())*0.001;
			request.objectmodify.position.y = ((double)recvlong())*0.001;
			request.objectmodify.position.z = ((double)recvlong())*0.001;
			request.objectmodify.direction.x = ((double)recvlong())*0.001;
			request.objectmodify.direction.y = ((double)recvlong())*0.001;
			request.objectmodify.direction.z = ((double)recvlong())*0.001;
			request.objectmodify.angle = ((double)recvlong())*0.001;
			request.objectmodify.timestamp = recvshort();
			break;

		  case UL_OBJECTMOVE:
			if ( incomplete( 22, 0 ) ) 
				break;
			if ( general.modifyobject == False ) {
				resp.response = UL_SMESSAGE;
				resp.smessage.to = fd;
				(const)resp.smessage.message = noobjmodify;
				netsend( fd, &resp );
				request.request = UL_FOO;
			}
			request.objectmove.id = recvlong();
			request.objectmove.dx = recvshort();
			request.objectmove.dy = recvshort();
			request.objectmove.dz = recvshort();
			request.objectmove.dwx = recvshort();
			request.objectmove.dwy = recvshort();
			request.objectmove.dwz = recvshort();
			request.objectmove.dangle = recvshort();
			request.objectmove.timestamp = recvshort();
			return &request;

		  case UL_SPECIAL:
			if ( incomplete( 5, 1 ) )
				break;
			request.special.from = recvshort();
			request.special.to = recvshort();
			request.special.type = recvq[recvused++];
			request.special.data = &recvq[recvused];
			while (recvq[recvused++]) { }
			return &request;

		  default:
			fprintf( stderr, "Unknown request %d; ignoring.\n",
				  request.request );
		  	request.request = UL_FOO;
			return &request;
		}
	
		/* If we get here, then we had an incomplete message at the
		 * end of recvq[].  Try to read the rest of it.
		 */
		memcpy(recvq, &recvq[beginning], recvcnt - beginning);
		recvcnt -= beginning;
		beginning = recvused = 0;
		i = netread(fd, &recvq[recvcnt], RECVQ - recvcnt);
		if (i <= 0)
			return NULL;
		recvcnt += beginning;
	}
/*NOTREACHED*/
}


/* Return TRUE if there is at least a partial message in the receive queue */
int netmore ( void ) {
	return (recvused < recvcnt);
}
