/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/

/* This file, netlisten.c, is adapted from original code by Steve Kirkendall.
Permission given. */

/* netlisten(int port)	returns a socket that listens on a given port number
 * netaccept(int lsock)	returns a socket that connects to a client
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <signal.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#define ADDRESS	struct sockaddr_in

/* Create a socket that listens on a given port */
int netlisten ( int port ) {
	struct sockaddr_in server;
	int	sock;

	/* Create the request socket */
	sock = socket(PF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		perror( "socket" );
		exit(1);
	}

	/* Give the socket a port number */
	server.sin_family = PF_INET;
	server.sin_port = htons((short)port);
	server.sin_addr.s_addr = INADDR_ANY;
	if ( bind(sock, (struct sockaddr *)&server, sizeof(server)) < 0 ) {
		perror( "bind socket" );
		exit(1);
	}

	/* Allow a backlog of up to 5 requests */
	if ( listen( sock, 5 ) < 0 ) {
		perror("listen");
		exit(1);
	}

#ifdef SIGPIPE
	/* Ignore SIGPIPE -- sockets are a variety of pipe, and we don't want
	 * the server to die if a client fails to read the response.
	 */
	signal(SIGPIPE, SIG_IGN);
#endif

	/* return the listening socket */
	return sock;
}


/* Create a socket that connects to the next client, given a listening socket.
 * It is assumed that either the listening socket already has a connection
 * request queued, or that the server should wait forever for one to arrive.
 * Returns the connected socket, or -1 if error.
 */
int netaccept ( int lsock ) {
	int		sock;		/* socket to client */
	ADDRESS		client_address;	/* address of client (ignored) */
	int		len;		/* length of address (ignored) */


	/* Accept the connection.  This creates a socket */
	len = sizeof(ADDRESS);
	sock = accept( lsock, (struct sockaddr *)&client_address, &len );
	if ( sock < 0 ) {
		perror( "could not accept connection" );
		return -1;
	}

	return sock;
}
