/* 
Unlimited Simulator is copyrighted (c) Bruno Barberi Gnecco, 1999, and
distributed under the GNU General Public License (GPL). The conditions can be
found in the file GPL, resident at the main directory of the distribution.
*/

/* This file, netselect.c, is adapted from original code by Steve Kirkendall.
Permission given. */

/* netaddsock(int sock)	Add a socket to the select list
 * netdelsock(int sock)	Delete a socket from the select list
 * netselect(long msec)	Wait up to msec milliseconds for data, return its socket
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#define MAXSOCK	256

static int	nsocks;		/* number of sockets to listen to */
static int	highest = -1;	/* highest socket number */
static int	list[MAXSOCK];	/* list of all sockets to listen to */
static fd_set	all;		/* bitmap of all sockets to listen to */
static fd_set	rfds;		/* bitmap of sockets with available data */
static int	gotfds;		/* number of sockets with data */

/* add a socket to the list of sockets to check */
void netaddsock ( int sock ) {
	/* set its bit and append it to the list */
	FD_SET(sock, &all);
	list[nsocks++] = sock;

	/* if this is the highest, then remember that */
	if (sock > highest)
		highest = sock;
}

/* remove a socket from the list of sockets to check */
void netdelsock ( int sock ) {
	int	i;

	/* turn off its bit */
	FD_CLR(sock, &all);

	/* in the list, replace it with the last item and reduce list size */
	for (i = 0; list[i] != sock; i++)
	{
	}
	list[i] = list[--nsocks];

	/* if it was highest, then find new highest */
	if (sock == highest)
	{
		for (highest = -1, i = 0; i < nsocks; i++)
			if (list[i] > highest)
				highest = list[i];
	}
}

/* Wait for available data (or an EOF condition) on any socket and return the
 * FD of one of them.  If timeout, then return -1.   Timeouts can be disabled
 * by passing -1 for msec.
 */
int netwait ( long msec ) {
	struct timeval	tv;	/* timeout value */
	int		i;

	/* if no data from last select() then wait again... */
	if (gotfds <= 0) {
		rfds = all;
		if (msec >= 0) {
			tv.tv_sec = msec / 1000;
			tv.tv_usec = (msec % 1000) * 1000;
			gotfds = select(highest + 1, &rfds, NULL, NULL, &tv);
		}
		else
			gotfds = select(highest + 1, &rfds, NULL, NULL, NULL);
		if (gotfds <= 0)
			return -1;
	}

	/* return one of the FDs that has data available */
	for (i = 0; !FD_ISSET(list[i], &rfds); i++) {
	}
	FD_CLR(list[i], &rfds);
	gotfds--;
	return list[i];
}
